import collections
import getpass
import functools
import logging
import logging.config

import IPython
import argh
import lxml
import traitlets

import nav
import nav.utils
from nav.wrappers import json


def _set_log_level(log_level):
    if log_level is not None:
        level = getattr(logging, log_level.upper())
        logging.basicConfig()
        logging.getLogger('zeep').setLevel(level)
        nav.logger.setLevel(level)


def _get_username(config_getter, username):
    return username or config_getter('username', None) or input('Username: ')


def _get_password(config_getter, password):
    return (
        password or
        config_getter('password', None) or
        getpass.getpass('Password: ')
    )


def meta(
    endpoint_type: 'Web services endpoint type',
    service_name: 'Web services endpoint',
    base_url: 'The base URL for the endpoint.' = None,
    username: 'Web services username' = None,
    password: 'Web services password' = None,
    log_level: 'The log level to use' = None,
    config_section: 'The config section to get settings from.' = 'nav',
):
    """Print the definition of a Codeunit or a Page"""
    _set_log_level(log_level)
    c = functools.partial(nav.config.get, config_section)
    username = _get_username(c, username)
    password = _get_password(c, password)
    data = nav.meta(
        endpoint_type=endpoint_type,
        service_name=service_name,
        base_url=c('base_url', base_url),
        username=username,
        password=password,
    )
    return lxml.etree.tostring(data, pretty_print=True).decode()


@argh.arg('-t', '--endpoint-type', type=str)
@argh.arg('-e', '--endpoint', type=str)
def interact(
    endpoint_type: 'Web services endpoint type' = None,
    endpoint: 'Web services endpoint' = None,
    base_url: 'The base URL for the endpoint.' = None,
    username: 'Web services username' = None,
    password: 'Web services password' = None,
    log_level: 'The log level to use' = None,
    config_section: 'The config section to get settings from.' = 'nav',
):
    """Starts a REPL to enable live interaction with a WSDL endpoint"""
    _set_log_level(log_level)
    c = functools.partial(nav.config.get, config_section)
    base_url = c('base_url', base_url)
    username = _get_username(c, username)
    password = _get_password(c, password)

    banner1_tmpl = """Welcome to nav client interactive mode
Available vars:
    {additional_arg}
    `create_service` - Create a WS service to make calls to

Example usage:
    service = create_service('Page', 'ItemList')
    results = service.ReadMultiple(filter=dict(No='123'), setSize=0)

    {additional_arg_example}
"""

    def create_service(endpoint_type, service_name):
        return nav.service(
            base_url,
            username,
            password,
            endpoint_type,
            service_name,
        )

    user_ns = {
        'create_service': create_service,
    }

    if endpoint_type and endpoint:
        banner1 = banner1_tmpl.format(
            additional_arg='`{0}` - Service to run WS calls against'.format(endpoint.lower()),
            additional_arg_example="{0}_results = {0}.ReadMultiple(filter=dict(Field='MyField', Criteria='My criteria'), setSize=0)".format(endpoint.lower())
        )
        user_ns[endpoint.lower()] = create_service(endpoint_type, endpoint)
    else:
        banner1 = banner1_tmpl.format(
            additional_arg='',
            additional_arg_example='',
        )

    IPython.embed(
        user_ns=user_ns,
        banner1=banner1,
        config=traitlets.config.Config(colors='LightBG')
    )


@argh.arg('-f', '--func-args', nargs='+', type=str)
def codeunit(
    service_name: 'Name of the code unit',
    func: 'Name of the function to run',
    base_url: 'The base URL for the endpoint.' = None,
    username: 'Web services username' = None,
    password: 'Web services password' = None,
    func_args: 'Add these kw args to the codeunit function call' = (),
    log_level: 'The log level to use' = None,
    config_section: 'The config section to get settings from.' = 'nav',
):
    """Get a Codeunit's results"""
    _set_log_level(log_level)
    c = functools.partial(nav.config.get, config_section)
    username = _get_username(c, username)
    password = _get_password(c, password)
    data = nav.codeunit(
        base_url=c('base_url', base_url),
        username=username,
        password=password,
        service_name=service_name,
        function=func,
        func_args=nav.utils.convert_string_filter_values(
            collections.OrderedDict(f.split('=') for f in func_args)
        ),
    )
    return json.dumps(data, indent=2)


@argh.arg('-f', '--filters', nargs='+', type=str)
@argh.arg('-n', '--num-results')
@argh.arg('-e', '--entries', nargs='+', type=str)
@argh.arg('-a', '--additional_data', nargs='+', type=str)
def page(
    service_name: 'Name of the WS page',
    func: 'Name of the function to use',
    base_url: 'The base URL for the endpoint.' = None,
    username: 'Web services username' = None,
    password: 'Web services password' = None,
    filters: 'Filters to apply to a ReadMultiple query' = (),
    entries: 'Entries to create when function is CreateMultiple' = (),
    additional_data: 'Additional data to pass alongside the main entries to create with CreateMultiple' = (),
    num_results: 'Amount of results to return' = 0,
    log_level: 'The log level to use' = None,
    config_section: 'The config section to get settings from.' = 'nav',
):
    """Get a Page's results"""
    _set_log_level(log_level)
    c = functools.partial(nav.config.get, config_section)
    username = _get_username(c, username)
    password = _get_password(c, password)

    data = nav.page(
        base_url=c('base_url', base_url),
        username=username,
        password=password,
        service_name=service_name,
        function=func,
        filters=nav.utils.convert_string_filter_values(
            collections.OrderedDict(f.split('=') for f in filters)
        ),
        entries=[
            nav.utils.convert_string_filter_values(
                collections.OrderedDict(
                    field.split('=') for field in entry.split(',')
                )
            )
            for entry in entries
        ],
        additional_data=nav.utils.convert_string_filter_values(
            collections.OrderedDict(
                ad.split('=') for ad in additional_data
            )
        ),
        num_results=num_results,
    )
    return json.dumps(data, indent=2)


command_parser = argh.ArghParser()
command_parser.add_commands([
    interact,
    meta,
    codeunit,
    page,
])
main = command_parser.dispatch



"""
Simple wrapper that adds some extra encoding capabilities needed for
this project.
"""
import datetime
import decimal
from json import JSONDecodeError  # noqa
import json as json_impl


class JsonExtendedEncoder(json_impl.JSONEncoder):
    """
    Needed for the json module to understand what to do with the types we
    use in this project.
    """
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, (
            datetime.time,
            datetime.date,
            datetime.datetime,
        )):
            return str(obj)
        return super().default(obj)


def dump(*args, **kw):
    if 'cls' not in kw:
        kw['cls'] = JsonExtendedEncoder
    return json_impl.dump(*args, **kw)


def dumps(*args, **kw):
    if 'cls' not in kw:
        kw['cls'] = JsonExtendedEncoder
    return json_impl.dumps(*args, **kw)


load = json_impl.load
loads = json_impl.loads



"""Conveniently make requests to Microsoft Dynamics NAV Web Services

Tested with Dynamics NAV 2009 R2. A lot of assumptions have been made while
making this project, so it cannot be guaranteed that it will work for everyone.

# Endpoint URL format is::

    http://<Server>:<WebServicePort>/<ServerInstance>/WS/<CompanyName>/<EndpointType>/<EndpointName>

More info here::

    https://msdn.microsoft.com/en-us/library/dd355398.aspx
"""
import collections
import logging

import requests
import requests_ntlm
import zeep
import zeep.cache

from . import config  # noqa
from . import exceptions
from ._metadata import __version__, __version_info__  # noqa
from .constants import (
    DEFAULT_WSDL_CACHE_EXPIRATION,
    CODEUNIT,
    PAGE,
    ReadMultiple,
    CreateMultiple,
)

logger = logging.getLogger('nav')


class NAV:

    def __init__(
        self,
        base_url: 'The base URL for the NAV web service.',
        username: 'Username (usually includes AD domain)',
        password: 'Password',
        cache_expiration: 'How long WSDL files are cached in memory. Set to something Falsy like False/0/None to disable this functionality.' = DEFAULT_WSDL_CACHE_EXPIRATION,
    ):
        self.base_url = base_url
        self.username = username
        self.password = password
        self.cache_expiration = cache_expiration
        self._service_cache = {}

    @staticmethod
    def _zeep_object_to_builtin_types(data, default=False):
        d = zeep.helpers.serialize_object(data)
        if d is None and default is not False:
            return default
        return d

    @staticmethod
    def _make_page_filters(filters):
        return [
            collections.OrderedDict([
                ('Field', field),
                ('Criteria', criteria),
            ])
            for field, criteria in filters.items()
        ]

    def _make_endpoint_url(self, *args):
        if self.base_url.endswith('/'):
            self.base_url = self.base_url[:-1]
        return '/'.join([self.base_url, *args])

    @staticmethod
    def _make_binding(endpoint_type, service_name):
        if endpoint_type == PAGE:
            urlpath_service_name = service_name.lower()
        else:
            urlpath_service_name = service_name
        return '{{urn:microsoft-dynamics-schemas/{0}/{1}}}{2}_Binding'.format(
            endpoint_type.lower(),
            urlpath_service_name,
            service_name,
        )

    def _run_capture_500(self, fun, *args, **kw):
        try:
            return fun(*args, **kw)
        except requests.exceptions.HTTPError as exc:
            raise exceptions.NAVHTTPError(
                *exc.args, request=exc.request, response=exc.response,
            )

    @staticmethod
    def validate_service_type(s):
        allowed_values = (CODEUNIT, PAGE)
        if s not in allowed_values:
            raise exceptions.InvalidServiceType(
                '`{}` is not a valid service type, must be one of {}'
                .format(s, allowed_values)
            )

    @staticmethod
    def validate_supported_page_function(s):
        allowed_values = (ReadMultiple, CreateMultiple)
        if s not in allowed_values:
            raise exceptions.UnsupportedPageFunction(
                '`{}` is not a supported service function, must be one of {}'
                .format(s, allowed_values)
            )

    def _make_client(self, endpoint_type, service_name):
        self.validate_service_type(endpoint_type)
        url = self._make_endpoint_url(endpoint_type, service_name)
        if self.cache_expiration:
            cache = zeep.cache.InMemoryCache(timeout=self.cache_expiration)
        else:
            cache = None

        session = requests.Session()
        session.auth = requests_ntlm.HttpNtlmAuth(self.username, self.password)
        transport = zeep.transports.Transport(session=session, cache=cache)

        return self._run_capture_500(
            zeep.Client, url, transport=transport, strict=False
        )

    def _make_service(
        self,
        endpoint_type: 'The endpoint type ("Page" or "Codeunit")',
        service_name: 'Name of the page/codeunit',
    ):
        """Initiate a WSDL service"""
        binding = self._make_binding(endpoint_type, service_name)
        if binding in self._service_cache:
            srvc = self._service_cache[binding]
        else:
            client = self._make_client(endpoint_type, service_name)
            srvc = client.create_service(binding, client.wsdl.location)
            self._service_cache[binding] = srvc
        return srvc

    def meta(
        self,
        endpoint_type: 'The endpoint type ("Page" or "Codeunit")',
        service_name: 'Name of the page/codeunit',
    ):
        """Get the definition of Codeunit or a Page"""
        client = self._make_client(
            endpoint_type,
            service_name,
        )
        return client.wsdl._get_xml_document(client.wsdl.location)

    def codeunit(
        self,
        service_name: 'The name of the code unit to use',
        function: 'Name of the code unit function to run',
        func_args: 'Add these kw args to the codeunit function call' = None,
    ):
        """Get a Codeunit's results"""
        srvc = self._make_service(
            endpoint_type=CODEUNIT,
            service_name=service_name,
        )
        func = getattr(srvc, function)
        data = func(**func_args)

        return self._zeep_object_to_builtin_types(data, default=[])

    def page(
        self,
        service_name: 'The name of the WS page',
        function: 'The function to use. Currently supported functions are ReadMultiple and CreateMultiple.',
        num_results: 'Maximum amount of results to return for ReadMultiple. Defaults to no limit.' = 0,
        filters: 'Apply filters to a ReadMultiple result' = None,
        entries: 'Entries to pass to CreateMultiple' = None,
        additional_data: 'Any additional data to pass along with the entries when using CreateMultiple.' = None,
    ):
        """Get a Page's results or create entries"""
        self.validate_supported_page_function(function)

        srvc = self._make_service(
            endpoint_type=PAGE,
            service_name=service_name,
        )

        if not filters:
            # NOTE: Workaround because the definition files for NAV 2009 R2 pages
            # requires the filter element to be defined (minOccurs=1), causing
            # zeep to raise a ValidationError exception. However, manually doing
            # an HTTP request to the web service, without passing in a filter
            # works. Perhaps create a ticket regarding a feature to temporarily
            # disable validation of min/maxOccurs.
            filters = {zeep.helpers.Nil(): zeep.helpers.Nil()}

        if function == ReadMultiple:
            data = srvc.ReadMultiple(
                filter=self._make_page_filters(filters),
                setSize=num_results,
            )
        elif function == CreateMultiple:
            if not entries:
                raise ValueError(
                    "Can't run Page CreateMultiple without passing in "
                    "any `entries`"
                )
            kw = dict(additional_data or {})
            kw.update({
                '{}_List'.format(service_name): [
                    {service_name: [entry for entry in entries]}
                ],
            })
            data = srvc.CreateMultiple(**kw)
        else:
            raise NotImplementedError

        return self._zeep_object_to_builtin_types(data, default=[])

    def read_multiple(
        self,
        service_name: 'The name of the WS page',
        num_results: 'Maximum amount of results to return. Defaults to no limit.' = 0,
        filters: 'Apply filters to the query' = None,
    ):
        return self.page(
            service_name=service_name,
            function=ReadMultiple,
            num_results=num_results,
            filters=filters,
        )

    def create_multiple(
        self,
        service_name: 'The name of the WS page',
        entries: 'Entries to pass to CreateMultiple' = None,
        additional_data: 'Any additional data to pass along with the entries when using CreateMultiple.' = None,
    ):
        return self.page(
            service_name=service_name,
            function=CreateMultiple,
            entries=entries,
            additional_data=additional_data,
        )


def _nav_factory(base_url, username, password, cache_expiration):
    return NAV(
        base_url=base_url,
        username=username,
        password=password,
        cache_expiration=cache_expiration,
    )


def meta(base_url, username, password, *args, **kw):
    return _nav_factory(
        base_url,
        username,
        password,
        cache_expiration=kw.pop(
            'cache_expiration',
            DEFAULT_WSDL_CACHE_EXPIRATION,
        ),
    ).meta(*args, **kw)


def service(base_url, username, password, *args, **kw):
    return _nav_factory(
        base_url,
        username,
        password,
        cache_expiration=kw.pop(
            'cache_expiration',
            DEFAULT_WSDL_CACHE_EXPIRATION,
        ),
    )._make_service(*args, **kw)


def page(base_url, username, password, *args, **kw):
    return _nav_factory(
        base_url,
        username,
        password,
        cache_expiration=kw.pop(
            'cache_expiration',
            DEFAULT_WSDL_CACHE_EXPIRATION,
        ),
    ).page(*args, **kw)


def codeunit(base_url, username, password, *args, **kw):
    return _nav_factory(
        base_url,
        username,
        password,
        cache_expiration=kw.pop(
            'cache_expiration',
            DEFAULT_WSDL_CACHE_EXPIRATION,
        ),
    ).codeunit(*args, **kw)



import collections

from . import constants


def convert_string_filter_values(filters):
    """Convert string type filter values to their XML equivalent
    Mostly useful for parsing CLI input data.
    """
    return collections.OrderedDict(
        (
            field,
            constants.STRING_VALUE_TO_PYTHON_TYPE_MAPPING.get(
                criteria,
                criteria
            )
        )
        for field, criteria in filters.items()
    )



from setuptools import setup

setup_kwargs = dict(
    name='nav',
    version='5.0.3',
    description='Conveniently make requests to Microsoft Dynamics NAV Web Services',
    packages=['nav', 'nav.wrappers'],
    include_package_data=True,
    author='Jacob Magnusson',
    author_email='m@jacobian.se',
    url='https://github.com/jmagnusson/nav',
    license='BSD',
    platforms='any',
    install_requires=[
        'lxml',
        'requests-ntlm',
        'zeep',
    ],
    extras_require={
        'cli': [
            'argh',
            'ipython',
        ],
        'test': {
            'coverage>=4.2',
            'flake8>=3.0.4',
            'pytest>=3.0.3',
            'responses>=0.5.1',
        },
    },
    entry_points={
        'console_scripts': [
            'nav = nav.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)

if __name__ == '__main__':
    setup(**setup_kwargs)



import requests
from lxml import etree


class InvalidServiceType(Exception):
    """Raised when an invalid service type is encountered

    I.e. not one of Page, Codeunit
    """


class UnsupportedPageFunction(Exception):
    """Raised when an invalid service function is encountered

    I.e. not one of ReadMultiple, CreateMultiple
    """


class NAVHTTPError(requests.exceptions.HTTPError):
    """Displays the error details that NAV returns"""

    def __str__(self):
        try:
            return ' - '.join(
                filter(
                    None,
                    (
                        e.text for e
                        in etree.fromstring(self.response.text).iter()
                        if e.tag in ('faultcode', 'faultstring', 'detail')
                    )
                )
            )
        except BaseException:
            return self.response.text
